import { DependencyManager } from "./dependency"
import "source-map-support/register"
import * as async from "async"
import * as fs from "fs"
import * as pathModule from "path"
import * as colors from "colors/safe"
import * as Leaf from "leaf-ts"
import { SharedCallbacks } from "./sharedCallbacks"
let useColors = colors

export class ServiceContext {
    services: {
        [index: string]: Service
    } = {}
    dependencyManager: DependencyManager = new DependencyManager;
    register(s: Service): void {
        this.dependencyManager.add(s.name, s.dependencies || null)
        this.services[s.name] = s
    }
    supportedLoadExtnames = [".ts", ".js"]
    loadServices(path: string) {
        path = pathModule.resolve(path)
        let files = fs.readdirSync(path)
        files = files.filter((item) => {
            let ext = pathModule.extname(item)
            if (this.supportedLoadExtnames.indexOf(ext) < 0) {
                return false
            }
            return true
        })
        files = files.map((item) => {
            let ext = pathModule.extname(item)
            let base = pathModule.basename(item, ext)
            return pathModule.join(path, base)
        })
        let toLoads = []
        for (let file of files) {
            if (toLoads.indexOf(file) >= 0) continue
            toLoads.push(file)
        }
        for (let file of toLoads) {
            this.loadService(file)
        }
    }
    loadService(path: string) {
        let result = require(path) || {}
        for (let name in result) {
            let obj = result[name]
            if (this.isServiceConstructor(obj)) {
                let Cons = obj as any
                let s = new Cons()
                if (!s.name) {
                    console.log(useColors.cyan(`Detect service without name. Use constructor name:${Cons.name}`))
                    s.name = Cons.name
                }
                console.log(`Load service ${useColors.green(Leaf.Util.capitalize(s.name))}`)
                //console.log(s,Cons)
                this.register(s)
            }
        }
    }
    private isServiceConstructor(obj) {
        return typeof obj === "function" && obj.prototype["thisObjectIsRootServiceConstructor"]
    }
    isSetup: boolean = false
    private finishSetup = SharedCallbacks.create()
    setup(finish: (err?: Error) => void = () => null): void {
        this.finishSetup.push(finish)
        if (this.isSetup) {
            return
        }
        this.isSetup = true
        let initialSequence = this.dependencyManager.resolve()
        async.eachSeries(initialSequence, (name, done) => {
            let service = this.services[name]
            service.services = this.services
            if (service.initialize) {
                if (service.initialize.length == 1) {
                    service.initialize((err?: Error) => {
                        if (err) {
                            console.error(useColors.red(`Fail to initialize service ${name}.`))
                            done(err)
                            return
                        }
                        this.logInitialized(name)
                        done()
                    })
                } else {
                    service.initialize()
                    this.logInitialized(name)
                    done()
                }
            } else {
                this.logInitialized(name)
                done()
            }
        }, (err: Error) => {
            if (err) {
                this.finishSetup(err)
                return
            }
            console.log(useColors.bgGreen("All service initialized"))
            this.finishSetup()
        })
    }
    private logInitialized(name: string) {
        console.log(`Service ${useColors.green(name)} Initialized`)
    }
}
export abstract class Service {
    abstract name: string
    services: {
        [index: string]: Service
    } = null
    dependencies: string[] = []
    isInitialized: boolean = false
    initialize(done?: (err?: Error) => void) {
        done()
    }
}
Service.prototype["thisObjectIsRootServiceConstructor"] = true
